{   PicBuild84 an assembler/disassembler for the PIC16F84 
 *  Version 1.06  August 1998
 *  FreeWare - not to be sold for profit
 *  Requires Turbo Pascal Version 6 or 7 to compile ( because of ASM )
 *  Update History:
        7/13/98 1.01 added '.' to 'F84' extention in writefile.
                     fixed so it will recognize TRIS instructions.
        7/15/98 1.02 fixed deleteproc so doesn't lock up when no instructions.
        7/16/98 1.03 added bit naming to nbr2op bit oriented types.
        7/18/98 1.04 added TRISA, TRISB and OPTION to instructions.
        8/03/98 1.05 added force line numbering in list. 
                    added F8 to insert current instruction before dest cursor.
                    fixed initial # for F9.
        8/04/98 1.06 added F10 to move dest cursor easily.
        8/06/98 1.07 fixed garbage showing up in bit instructions.
       10/26/98 1.08 added copy from present instruction to operation window.
                     added increment of cursors and copy of comments to F8.
}

{$V-}                                              { turn off string checking }

USES DOS;


TYPE Str80 = String[80];
     Str40 = String[40];
     Str6 = string[6];

CONST
    cursor1 : Integer =  $3E;                { cursor for operation selection }
    cursor2 : Integer =  $4F;              { cursor for instruction selection }
    cursor3 : Integer =  $5D;              { cursor for destination selection }
    Vbase : Word = $B800;                         { video base, default color }

                                            { forms various instructions take }
  keylabels : ARRAY [0..37] OF String[10]= (
'MOVLW k','ADDLW k','ANDLW k','CALL','XORLW k','CLRW','BCF f,b','INCF f,d',
'MOVWF f','ADDWF f,d','ANDWF f,d','RETURN','XORWF f,d','CLRF f','BSF f,b','INCFSZ f,d',
'MOVF f,d','SUBLW k','IORLW k','RETLW k','RLF f,d','COMF f,d','BTFSC f,b','DECF f,d',
'NOP','SUBWF f,d','IORWF f,d','RETFIE','RRF f,d','SWAPF f,d','BTFSS f,b','DECFSZ f,d',
'GOTO','SLEEP','CLRWDT','TRISA','TRISB','OPTION');

                                     { 'raw' opcodes for various instructions }
  opcode : ARRAY [0..37] OF Word = ( 
$3000,$3E00,$3900,$2000,$3a00,$0103,$1000,$0A00,
$0080,$0700,$0500,$0008,$0600,$0180,$1400,$0F00,
$0800,$3C00,$3800,$3400,$0D00,$0900,$1800,$0300,
$0000,$0200,$0400,$0009,$0C00,$0E00,$1C00,$0B00,
$2800,$0063,$0064,$0065,$0066,$0062);

                                                { sizes of instruction labels }
labelsize  : ARRAY [0..37] OF Byte = ( 
     7,7,7,4,7,4,7,8,
     7,9,9,6,9,6,7,10,
     8,7,7,7,7,8,9,8,
     3,9,9,6,7,9,9,10,
     4,5,6,5,5,6);

                               { names of special purpose & general registers }
reg : ARRAY [0..79] OF String[11] =(  
'INDF','TMR0/OPTION','PCL','STATUS','FSR',
'PORTA/TRISA','PORTB/TRISB','not used','EEDATA/CON1','EEADR/CON2',
'PCLATH','INTCON','REG00','REG01','REG02',
'REG03','REG04','REG05','REG06','REG07',
'REG08','REG09','REG10','REG11','REG12',
'REG13','REG14','REG15','REG16','REG17',
'REG18','REG19','REG20','REG21','REG22',
'REG23','REG24','REG25','REG26','REG27',
'REG28','REG29','REG30','REG31','REG32',
'REG33','REG34','REG35','REG36','REG37',
'REG38','REG39','REG40','REG41','REG42',
'REG43','REG44','REG45','REG46','REG47',
'REG48','REG49','REG50','REG51','REG52',
'REG53','REG54','REG55','REG56','REG57',
'REG58','REG59','REG60','REG61','REG62',
'REG63','REG64','REG65','REG66','REG67');

                 { ************* some constants used to describe various keys }
  LARR  = #203;                 { Left Arrow }
  RARR  = #205;                 { Right Arrow }
  UARR  = #200;                 { Up Arrow }
  DARR  = #208;                 { Down Arrow }
  ESC   = #27;                  { Escape Key }
  BACKSP = #8;                  { Backspace Key }
  CR    = #13;                  { Carriage Return }
  DEL = #211;                   { Delete Key }
  INS = #210;                   { Insert Key }
  ENTER = #13;                  { Enter Key }
  HOME  = #199;                 { Home Key }
  ENDKEY = #207;                { End Key }
  PGUP   = #201;                { Page Up Key }
  PGDN   = #209;                { Page Down Key }
  CtrlPgUp = #4;                { Ctrl - PgUp Key }
  CtrlPgDn = #246;              { Ctrl - PgDn Key }
  CtrlEnterKp = #10;            { Ctrl - Enter - Keypad Key }
  CtrlPlusKp = #16;             { Ctrl - Plus - Keypad Key }
  CtrlMinusKp = #14;            { Ctrl - Minus - Keypad Key }
  CtrlHome = #247;              { Ctrl - Home Key }
  CtrlEnd = #245;               { Ctrl - End Key }
  F3 = #189;                    { F3 key }

VAR
 lines : ARRAY [0..1023] OF Word;                 { holds instruction opcodes }
 edata : ARRAY [0..63] OF Word;                           { holds eeprom data }
 labels : ARRAY[0..1023] OF String[16];                  { label of each line }
 comments : ARRAY[0..1023] OF String[38];             { comment for each line }
 banks : ARRAY[0..1023] OF Boolean;                { RB0 status for each line }
 fname : String[8];                                           { for filenames }
 filestuff : SearchRec;                        { used in findfirst - findnext }
 ffirst : Boolean;                             { flag for sucessful findfirst }
 flnbr : Boolean;                              { used to force line numbering }


{**********************************************************************}

   PROCEDURE HideCursor; Assembler;
     ASM  mov ax,$0100; mov cx,$2607; int $10 END;

{**********************************************************************}

   PROCEDURE ShowCursor; Assembler;
     ASM  mov ax,$0100; mov cx,$0506; int $10 END;

{**********************************************************************}

   Function mode: Byte; assembler;   { returns current video mode }
         ASM mov ah,$f; int $10; xor ah,ah END;

{**********************************************************************}

   PROCEDURE ClrScr; Assembler;
     ASM mov ax,Vbase; mov es,ax; mov di,0; MOV cx,2000;
           MOV ax,$0F00 ; CLD; REP STOSW END;
 
{**********************************************************************}

  PROCEDURE fwrite(col,row:Integer;s:String;att:Integer); 
    VAR offset,j : Word;
        
    BEGIN
      offset := 160*Pred(row)+2*Pred(col);
      FOR j:=1 TO Length(s) DO
         BEGIN
           MemW[Vbase:offset] := (att SHL 8) + Ord(s[j]);
           Inc(offset,2);
         END
    END;                      

{**********************************************************************}

PROCEDURE ClearBox(x,y,w,h,att:Integer);
   CONST
   topline = ''+
             'Ŀ';
   botline = ''+
             '';
   VAR s : String[80];
       j,rs : Integer;
  BEGIN
    s:= topline;
    Delete(s,2,80-w);
    fwrite(x,y,s,att);
    rs:= x+Pred(w);
    FOR j:= 1 TO h-2 DO BEGIN
         fwrite(x,y+j,'',att); 
         fwrite(rs,y+j,'',att);
        END;
    s:= botline;
    Delete(s,2,80-w);
    fwrite(x,y+Pred(h),s,att);
  END;

{**********************************************************************}
  FUNCTION Pad2len(s:string;len:Byte):string;
    VAR p : Byte;
    BEGIN 
      FOR p := Length(s) TO len DO s := s + ' ';
      Pad2len := s;
    END;
{**********************************************************************}

FUNCTION itoa(n:integer):str6; assembler;
  ASM
            les di,@Result     { DI = string pointer }
            mov ax,n

            xor cx,cx          { Zero CX  }
            test ax,ax         { Check for zero }
            jnz @p1_nozero
            mov bx,'0'
            push bx            { Push a zero }
            inc cx             { One digit }
            jmp @p1_zloop

@p1_nozero: mov si,10          { SI = 10  }

@p1_dloop:  xor dx,dx          { Divide by 10 }
            div si
            mov bl,dl          { Remainder in BL  }
            add bl,30h         { Convert to digit }
            push bx            { Push digit }
            inc cx
            test ax,ax         { Loop back  }
            jnz @p1_dloop
@p1_zloop:  mov ax,cx          { length byte }
            stosb
@p1_ploop:  pop ax             { Pop digit  }
            stosb              { Store digit }
            loop @p1_ploop      { Loop back }

  END;


{**********************************************************************}

  PROCEDURE GotoXY(x,y:Byte);
            BEGIN Dec(x); Dec(y);
            asm mov ah,2; mov bh,0; mov dl,x; mov dh,y; int 10h; end;
            END;

{**********************************************************************}

      FUNCTION Getkey : Char;  { Get a key press add 128 to special keys }
        VAR key : Byte;
        BEGIN
           asm
              mov ah, 7  { read keyboard }
              int 21h    { using dos }
              cmp al, 0  { zero? }
              jne @1     { no, use it }
              int 21h    { read again }
              add al, 128  { special key }
          @1: mov key,al
            end;
          GetKey := Chr(key);
        END;    { Getkey }

{**********************************************************************}

  PROCEDURE ClearArea; Assembler;
   ASM mov ax,Vbase; mov es,ax; mov di,16*160; MOV cx,240;
           MOV ax,$0700 ; CLD; REP STOSW END;


 {**********************************************************************}

      PROCEDURE GetString(prompt:Str40;len:byte;VAR s:Str40;post:Str6);
        VAR clearit,worker : Str40;
            lw,curx,strtp,strtb : byte;
            ch : Char;
            firstchr : Boolean;
        BEGIN
            ShowCursor;
            ClearArea;
            strtp :=(78-Length(prompt)-len) DIV 2;
            strtb := strtp + Length(prompt) +1;
            fwrite(strtp,18,prompt,$E);     
            fwrite(strtb+len+1,18,post,$E);     
            Fillchar(clearit,SizeOf(clearit),'.');
            clearit[0] := Chr(len);
            worker := s;
            curx := length(worker) +1;
            if curx > len THEN curx := len;
            firstchr := TRUE;
            REPEAT
              fwrite(strtb+1,18,clearit,$F);
              fwrite(strtb+1,18,worker,$F);
              GotoXY(curx+strtb,18);
              ch := Getkey;
              lw := Length(worker);  
              CASE ch OF
                DEL    : IF curx <= lw THEN BEGIN
                              delete(worker,curx,1); 
                             { IF curx > 1 THEN Dec(curx);} END;
                BACKSP : IF curx >1 THEN BEGIN
                             delete(worker,curx-1,1);
                              IF curx > 1 THEN Dec(curx); END;
                LARR   : IF curx > 1 THEN Dec(curx);
                RARR   : IF (curx <= lw) AND (curx < len) THEN Inc(curx); 
                HOME   : curx := 1; 
                ENDKEY : curx := lw+1; 
                F3 : IF ffirst = TRUE THEN 
                        BEGIN
                          FindNext(filestuff);
                          worker := Copy(filestuff.name,1,Pred(Pos('.',filestuff.name)));
                        END;
            ' '..'}'    :  IF firstchr THEN BEGIN
                              worker := ch; curx := 2 END
                            ELSE
                             IF (lw < len) THEN
                               BEGIN        { room for at least one more }
                                 IF (curx <= lw) THEN  { in position for insert }
                                   BEGIN
                                     insert(ch,worker,curx);
                                     Inc(curx);
                                   END
                                   ELSE                      { add to end }
                                   BEGIN
                                     worker := worker +ch;
                                     Inc(curx);
                                     IF curx > len THEN curx := len;
                                   END;
                               END
                               ELSE
                                 IF curx = len THEN worker[len] := ch;
             END;   { Case }
             firstchr := FALSE;
           UNTIL ch IN [ESC,CR];
           IF ch = CR THEN s := worker;
           HideCursor;
      END;  { GetStr }

{**********************************************************************}

  Function Byte2Hex(numb : Byte): String;       { Converts Byte to hex String }
    Const
      HexChars : Array[0..15] of Char = '0123456789ABCDEF';
    begin
      Byte2Hex[0] := #2;
      Byte2Hex[1] := HexChars[numb shr  4];
      Byte2Hex[2] := HexChars[numb and 15];
    end; { Byte2Hex }

{**********************************************************************}

  Function Word2Hex(nbr : Word): String;       { Converts Word to hex String }
    begin
      Word2Hex := Byte2Hex(hi(nbr)) + Byte2Hex(lo(nbr));
    end;

{**********************************************************************}

   PROCEDURE CommentBox;
     BEGIN
       ClearArea;
       ClearBox(38,17,40,3,$E);                                { comment box }
       fwrite(56,17,'Comment',$E);
       fwrite(58,19,'F7',$E);
       ClearBox(1,17,18,3,$E);                                   { label box }
       fwrite(7,17,'Label',$E);
       fwrite(9,19,'F6',$E);
       fwrite(22,18,fname+'.HEX',$F);                             { filename }
      END;

{**********************************************************************}
   
   PROCEDURE Message(s:Str40);
   VAR ch : Char;
       strtm : Byte;
     BEGIN
       ClearArea;
       strtm := (78 - length(s)) DIV 2;
       ClearBox(Pred(strtm),17,length(s)+2,3,$4C);     
       fwrite(strtm,18,s,$4F);
       ch := Getkey;
       CommentBox;
     END;

{**********************************************************************}
   
   FUNCTION Response(s:Str80):Char;
   VAR ch : Char;
       strtm : Byte;
     BEGIN
       ClearArea;
       strtm := (78 - length(s)) DIV 2;
       ClearBox(Pred(strtm),17,length(s)+2,3,$1E);     
       fwrite(strtm,18,s,$4F);
       ch := Getkey;
       CommentBox;
       Response := ch;
     END;

{**********************************************************************}
   
   PROCEDURE Chgregname(nbr:Integer);
     BEGIN
       Getstring('Enter New Register Name',11,reg[nbr],'');
       CommentBox;
    END;

{**********************************************************************}
   
   PROCEDURE InsertLabel(nbr:Integer);
     BEGIN
       Getstring('Enter New Label For Line #'+itoa(nbr),16,labels[nbr],'');
       CommentBox;
     END;

{**********************************************************************}
   
   PROCEDURE InsertComment(nbr:Integer);
     BEGIN
       Getstring('Enter New Comment For Line #'+itoa(nbr),38,comments[nbr],'');
       CommentBox;
     END;

{**********************************************************************}

PROCEDURE helpscreen;                             { display help screen (F1) }
VAR ch : Char;
BEGIN                     
  ClrScr;
  ClearBox(1,1,80,25,$1B);
  fwrite(20,1,' PicBuild84 Help Screen Version 1.08 ',$1B);
  fwrite(6,3,'Home/End - program cursor',$0E);
  fwrite(6,4,'Enter - replace op at cursor',$0E);
  fwrite(6,5,'Insert - insert op before cursor',$0E);
  fwrite(6,6,'Delete - delete op at cursor',$0E);
  fwrite(6,7,'Ctrl Home/End destination cursor',$0E);
  fwrite(6,8,'Arrow keys - select operation',$0E);
  fwrite(6,9,'F6 - Label line at cursor',$0E);
  fwrite(6,10,'F7 - Comment line at cursor',$0E);
  fwrite(16,12,'Register Window',$0D);
  fwrite(6,13,'PgUp/PgDn - register cursor',$0D);
  fwrite(6,14,'F5 Rename register (>12)',$0D);
  fwrite(6,15,'Tab - toggle dest bit (f/W)',$0D);
  fwrite(6,20,'Ctrl PgUp/Dn - eedata cursor',$0A);
  fwrite(6,21,'Ctrl + (keypad) increment data',$0A);
  fwrite(6,22,'Ctrl - (keypad) decrement data',$0A);
  fwrite(6,23,'Ctrl Enter toggle selected bit in data',$0A);
  fwrite(6,24,'Ctrl left/right arrows select current bit',$0A);
  fwrite(6,17,'Alt 0-9 (top row) - toggle selected bit in k',$0B);
  fwrite(6,18,'Alt +/- (top row) inc/dec (k)',$0B);
  fwrite(12,2,'Program Keys:',$0F);
  fwrite(8,11,'Instruction Creation Keys:',$0F);
  fwrite(15,19,'Data Keys:',$0F);
  fwrite(18,16,'Literal Keys:',$0B);
  fwrite(40,18,'F11 - Copy current op to op window',$0E);
  fwrite(40,19,'F8 - Copy current op before dest cursor',$0E);
  fwrite(40,20,'F2 - Save .Hex and .F84 files to disk',$0E);
  fwrite(40,21,'F3 - Load files (repeat F3 for names)',$0E);
  fwrite(40,22,'F4 - List to Screen, Disk or Printer',$0E);
  fwrite(50,23,'F9 - Goto line #',$0E);
  fwrite(50,24,'F10 - Set destination',$0E);
  fwrite(39,2,'In this assembler you build each',$0F);
  fwrite(39,3,'instruction from given building blocks.',$0F);
  fwrite(39,4,'It appears in the box in the center of',$0F);
  fwrite(39,5,'the screen.  Once you have the',$0F);
  fwrite(39,6,'instruction you want, it is placed in',$0F);
  fwrite(39,7,'the program listed in the left hand box',$0F);
  fwrite(39,8,'on the screen. You may replace the',$0F);
  fwrite(39,9,'instruction under the cursor bar or',$0F);
  fwrite(39,10,'insert it at that position. On',$0F);
  fwrite(39,11,'insertion all calls and goto''s will',$0F);
  fwrite(39,12,'automatically be fixed.  Programs are',$0F);
  fwrite(39,13,'saved in INHX8M format, (XXX.HEX ).',$0F);
  fwrite(39,14,'At the same time a file XXX.F84 is saved',$0F);
  fwrite(39,15,'which holds labels etc.',$0F);
  Ch := GetKey;
END;

{**********************************************************************}

  { change attribute of nbr chars at col,row  } 
  PROCEDURE chgattr(col,row,nbr,attr:Integer); 
    VAR offset,j : Integer;
    BEGIN
      offset := 160*Pred(row)+2*Pred(col)+1;
      FOR j:=0 TO Pred(nbr) DO
         BEGIN
           Mem[Vbase:offset] := attr;
           Inc(offset,2);
         END
    END;                      

{**********************************************************************}
  TYPE Str26 = string[26];
  FUNCTION nbr2op(n:Word;b:Boolean):Str26; { print out operation given opcode }      
    CONST
    statusbits : ARRAY[0..5] OF String[3] = ('C','DC','Z','PD','TO','RP0');
    intconbits : ARRAY[0..7] OF String[4] =
      ('RBIF','INTF','T0IF','RBIE','INTE','T0IE','EEIE','GIE');
    optionbits : ARRAY[0..7] OF String[6] =
      ('PS0','PS1','PS2','PSA','T0SE','T0CS','INTEDG','RBPU');
    eecon1bits : ARRAY[0..4] OF String[5] = ('RD','WR','WREN','WRERR','EEIF');
     VAR
       s : Str26; bitstr : String[6];
       temp : String[11];
     BEGIN
       s :='?????';
       temp := reg[n AND $7F];
       CASE n AND $7F OF
          1 : IF b = TRUE THEN temp := 'OPTION' ELSE temp := 'TMR0';
          5 : IF b = TRUE THEN temp := 'TRISA' ELSE temp := 'PORTA';
          6 : IF b = TRUE THEN temp := 'TRISB' ELSE temp := 'PORTB';
          8 : IF b = TRUE THEN temp := 'EECON1' ELSE temp := 'EEDATA';
          9 : IF b = TRUE THEN temp := 'EECON2' ELSE temp := 'EEADR';
       END;  { case }
       CASE (n SHR 12) OF
         3: BEGIN                                 { Literal operations }
           CASE (n SHR 8 AND $0F) OF
            0: s :='MOVLW';
            4: s := 'RETLW';
            8: s := 'IORLW';
            9: s := 'ANDLW';
            $0A: s :='XORLW';
            $0C: s :='SUBLW';
            $0E: s :='ADDLW';
           END;
          s := s + ' $' + Byte2Hex(n AND $0FF);
         END;
        2: BEGIN                                 { goto or call }
            IF (n AND $0800 > 0 ) THEN s :='GOTO'
              ELSE s :='CALL';
            s := s + ' ' + labels[n AND $07FF]+' ('+ itoa(n AND $07FF)+')';
          END;
        1: BEGIN                            { bit oriented }
             bitstr := itoa(n SHR 7 AND 7);
             CASE (n AND $7F) OF
               1: IF b THEN bitstr := optionbits[(n SHR 7) AND 7];
               3: bitstr := statusbits[(n SHR 7) AND 7];
               8: IF b THEN bitstr := eecon1bits[(n SHR 7) AND 7];
              $B: bitstr := intconbits[(n SHR 7) AND 7];
             END;  { Case }
             CASE (n SHR 10 AND 3) OF
               0: s :='BCF';
               1: s :='BSF';
               2: s :='BTFSC';
               3: s :='BTFSS';
             END;  { Case }
            s := s + ' ' + temp +','+bitstr;
           END;
        0: BEGIN                          { register oriented }
           CASE (n SHR 8 AND $0F) OF
            0: if ( n SHR 7 AND 1) > 0 THEN s :='MOVWF'
                    ELSE CASE(n AND $0F) OF
                         0: s :='NOP';
                         2: s :='OPTION';
                         3: s :='SLEEP';
                         4: s :='CLRWDT';
                       5..6: s :='TRIS';
                         8: s :='RETURN';
                         9: s :='RETFIE';
                     END;
            1: IF ( n SHR 7 AND 1)> 0 THEN s:='CLRF'
                    ELSE s :='CLRW';
            2: s :='SUBWF';
            3: s :='DECF';
            4: s :='IORWF';
            5: s :='ANDWF';
            6: s :='XORWF';                      
            7: s :='ADDWF';
            8: s :='MOVF';
            9: s :='COMF';
            $0A: s :='INCF';
            $0B: s :='DECFSZ';
            $0C: s :='RRF';
            $0D: s :='RLF';
            $0E: s :='SWAPF';
            $0F: s :='INCFSZ';
           END;
       IF ((n SHR 8 AND $0F) > 1) THEN   { not clrf,clrw,movwf or nop }
         BEGIN
            s := s + ' ' + temp;
            IF(n SHR 7 AND 1) > 0 THEN  s := s +',f'     { add d bit }
               ELSE s := s +',W';
         END
         ELSE IF (n SHR 7 AND 1 ) > 0 THEN         { has register assocated }
               s := s + ' ' + temp;
       IF n = $65 THEN s := s + ' TRISA';
       IF n = $66 THEN s := s + ' TRISB';
      END;  { 0 }
    END;  { Case }
    nbr2op := s;
  END;  { nbr2op }              

{**********************************************************************}

     
PROCEDURE readfile;               { input file with title 'name' }
  VAR
    savename : string[8];
    infile : TEXT;
    s : Str80; temp : string[5];
    j,w,size,addr,code,p: Word;
  LABEL getcomments;
  BEGIN
    savename := fname;               
    FindFirst('*.hex',$10,filestuff);
    IF DosError = 0 THEN 
       BEGIN
         fname := Copy(filestuff.name,1,Pred(Pos('.',filestuff.name)));
         ffirst := TRUE
       END;
    Getstring('Enter Name of File To Read',8,fname,'.HEX');
    Assign(infile,fname+'.HEX');
    {$I-}  reset(infile);  {$I+}
    IF IOResult > 0 THEN
       BEGIN
         message('Couldn''t open '+ fname +'.HEX!');
         fname := savename;
         Exit;
       END;
    FOR j:=0 TO 1023 DO lines[j] := 0;                 { zero out old program }
    FOR j:=0 TO 1023 DO labels[j] := '';
    FOR j:=0 TO 1023 DO comments[j] := '';          
    FOR j:=0 TO 63 DO edata[j] := 0;                               { and data }
    WHILE NOT Eof(infile) DO                          { get file line by line }
      BEGIN
        Readln(infile,s);
        IF s[9] <> '0' THEN goto getcomments;
        temp := '$'+ Copy(s,2,2);                  { get byte count into size }
        val(temp,size,code);
        temp := '$' + Copy(s,4,4);                    { get address into addr }
        val(temp,addr,code);
        addr := addr DIV 2;          { change address from byte to word value }
        IF (addr < 1024) THEN                   { if program instructions ... }
           FOR j:=0 TO Pred(size DIV 2) DO                       { get word by word }
             BEGIN
              temp := '$' + Copy(s,10+4*j,4);
              val(temp,w,code);
              w := Swap(w);                              { swap hi & lo bytes }
              lines[addr+j] := w;                    { save as program opcode }
             END;
        IF (addr >= $2100) THEN                              { if eeprom data }
           FOR j:=0 TO Pred(size DIV 2) DO                       { get word by word }
             BEGIN
              temp := '$' + Copy(s,10+4*j,4);
              val(temp,w,code);
              w := w SHR 8;                             { save hi  byte .... }
              edata[addr-$2100+j] := w;                            { as data }
             END;
       END;  { while } 
   getcomments:
     Close(infile);
     Assign(infile,fname+'.F84');
     {$I-}  reset(infile);  {$I+}
     IF IOResult > 0 THEN
       BEGIN
         message('Couldn''t open '+ fname+'.F84');
         Exit;
       END;
     WHILE NOT Eof(infile) DO
      BEGIN 
        Readln(infile,s);
        p := Pos(' ',s);
        IF p > 0 THEN temp := Copy(s,1,Pred(p));
        Val(temp,j,code);
        Delete(s,1,p);                                   { get rid of number }
        IF j < 1024 THEN                             { an instruction number }
          BEGIN                                                                
            p := Pos(';',s);
            IF p > 1 THEN
              labels[j] := Copy(s,1,Pred(p));
            Delete(s,1,p);                                { get rid of label }
            IF s <> '' THEN comments[j] := s;
          END
        ELSE 
          reg[j-2000] := s;               { register numbers have 2000 added }
      END;
     Close(infile);
    END;

{**********************************************************************}

  PROCEDURE writefile;                    { output program to file of 'name' }
  VAR
    outfile : TEXT;
    s : Str80; temp : string[5];
    addr,diff,j,last,chksum,p: Integer;
  LABEL writedata,closefile;
  BEGIN
    Getstring('Enter Name of File To Write',8,fname,'.HEX');
    Assign(outfile,fname+'.HEX');
    {$I-}  rewrite(outfile);  {$I+}
    IF IOResult > 0 THEN
       BEGIN
         message('Couldn''t open '+ fname +'.HEX!');
         Exit;
       END;
     addr := 0; j := 1023;                            { find last instruction }
     While(lines[j] = 0) DO Dec(j);
     IF j < 0 THEN GOTO writedata;
     last := j;
     WHILE (addr < last +1) DO                      { first write out program }
       BEGIN
         chksum := 0;
         diff := last - addr + 1;
         IF (diff > 8) THEN diff := 8;                { 8 words at a time max }
         Write(outfile,':',Byte2Hex(diff*2)); { save ':' + byte count to file }
         chksum := chksum + diff*2;                 { bytes to save to chksum }
         write(outfile,Word2Hex(addr*2));              { address word to file }
         chksum := chksum + (addr*2) AND $FF;   { low & high bytes of address }
         chksum := chksum + (addr*2) SHR 8;                       { to chksum }
         write(outfile,'00');
         FOR j:=0 TO Pred(diff) DO                    { for each program word }
           BEGIN
              write(outfile,Byte2Hex(lines[addr] AND $FF));    { save lo byte }
              chksum := chksum + lines[addr] AND $FF;      { add it to chksum }
              write(outfile,Byte2Hex(lines[addr] SHR 8));      { then hi byte }
              chksum := chksum + lines[addr] SHR 8;
              Inc(addr);
           END;
          Dec(chksum);                            { twos compliment of chksum }
          chksum :=  chksum XOR $FF;
          chksum := chksum AND $FF;
          writeln(outfile,Byte2Hex(chksum));                    { saved to file }
      END;
     writedata:                                             { now output data }
     addr := $2100; j := 63;
     WHILE edata[j] = 0 DO Dec(j);                   { find last nonzero data }
     IF j < 0 THEN GOTO closefile;                                     { none }
     last := $2100 + j;
     WHILE addr < last + 1 DO
       BEGIN
         chksum := 0;
         diff := last - addr + 1;
         IF (diff > 8) THEN diff := 8;                { 8 words at a time max }
         Write(outfile,':',Byte2Hex(diff*2)); { save ':' + byte count to file }
         chksum := chksum + diff*2;                 { bytes to save to chksum }
         chksum := chksum + (addr*2) AND $FF;   { low & high bytes of address }
         chksum := chksum + (addr*2) SHR 8;                       { to chksum }
         write(outfile,Word2Hex(addr*2));                    { address to file}
         write(outfile,'00');
         FOR j:=0 TO Pred(diff) DO                    { for each program word }
           BEGIN
              write(outfile,Byte2Hex(edata[addr-$2100] AND $FF)); {save lo byte}
              chksum := chksum + edata[addr-$2100] AND $FF;{ add it to chksum }
              write(outfile,'00');                        { all hi bytes := 0 }
              Inc(addr);
           END;
          Dec(chksum);                            { twos compliment of chksum }
          chksum := chksum XOR $FF;
          chksum := chksum AND $FF;
          writeln(outfile,Byte2Hex(chksum));                    { saved to file }
     END;
     closefile:
     write(outfile,':00000001FF');
     close(outfile);
    Assign(outfile,fname+'.F84');
    {$I-}  rewrite(outfile);  {$I+}
    IF IOResult > 0 THEN
       BEGIN
         message('Couldn''t open '+ fname+'.F84!');
         Exit;
       END;
    FOR j:= 0 TO 1023 DO
      IF (labels[j] <> '') OR (comments[j] <> '') THEN
        Writeln(outfile,itoa(j)+' '+labels[j]+';'+comments[j]);
    FOR j:= 12 TO 67 DO
      IF Copy(reg[j],1,3) <> 'REG' THEN
        Writeln(outfile,itoa(j+2000)+' '+reg[j]);
    Close(outfile);
   END;

{**********************************************************************}

  FUNCTION insertproc(cur:Integer): Integer;  { insert operation into program }
    VAR  j : Integer;
         ch : Char;
    BEGIN
      IF (lines[1023] > 0) THEN                { check for room for insertion }
         BEGIN
             message('There is not enough room!');
             insertproc := 0;
             Exit;
         END;
      { move cur & above up }
    FOR j := 1023 DOWNTO Succ(cur) DO lines[j] := lines[j-1]; 
    FOR j := 1023 DOWNTO Succ(cur) DO labels[j] := labels[j-1]; 
    FOR j := 1023 DOWNTO Succ(cur) DO comments[j] := comments[j-1]; 
    FOR j := 0 TO 1023 DO                           { check all instructions }
     { all call & gotos have destination incremented if the destination
            is beyond or equal to the insertion location }
           IF ((lines[j] AND $3000) = $2000) AND
              ((lines[j] AND $07FF) >= cur) THEN Inc(lines[j]); 
                                          { changed, incr. }
    labels[cur] := '';
    comments[cur] := '';
    insertproc := 1; 
    END;

{**********************************************************************}

  PROCEDURE deleteproc(cur:Integer);         { delete operation from program }
    VAR  j, last : Integer;
    BEGIN
    last := 0;
    FOR j:=1023 DOWNTO 0 DO
          IF (lines[j] > 0) OR (labels[j] <>'') OR (comments[j]<>'')  THEN
             BEGIN  last := j; j:= 0; END;            { find last instruction }
    FOR j:=cur TO Pred(last) DO lines[j] := lines[j+1]; { move above cur down }
    FOR j:=cur TO Pred(last) DO labels[j] := labels[j+1]; 
    FOR j:=cur TO Pred(last) DO comments[j] := comments[j+1]; 
    FOR j := 0 TO 1023 DO                            { check all instructions }
           IF ((lines[j] AND $3000) = $2000 ) THEN  { if call or goto then .. }
              IF ((lines[j] AND $07FF) > cur) THEN Dec(lines[j]);
                              { changed, decr. }
    lines[last] := 0;                                  { NOP in last position }
    labels[last] :='';
    comments[last] :='';
    END; 

{**********************************************************************}

  PROCEDURE prtbin(nbr:Integer);
    VAR  mask,j: Integer;
    BEGIN
      mask := 128;
      FOR j:=0 TO 7 DO                             { print out binary version }
        BEGIN
          IF ( mask AND nbr ) > 0 THEN Write('1 ')
            ELSE Write('0 ');
          mask :=  mask SHR 1;
        END;
       Write(nbr:3,' $',Byte2Hex(nbr),' ');            { dec, ascii versions }
       IF not(nbr IN [7,9]) THEN Write(Chr(nbr));     { bell and tab screw up }
    END;

{**********************************************************************}

   PROCEDURE ListFile(name:Str40);
     VAR F : TEXT;
         j,last,p,row : Integer;
         temp : Str26;
         ch : Char;
     LABEL out1;
     BEGIN
       Assign(F,name);
       {$I-}  rewrite(F);  {$I+}
       IF IOResult > 0 THEN
         BEGIN
              message('Couldn''t open '+ name +'!');
              Exit;
         END;
       j := 1023;                               { find the last instruction }
       WHILE lines[j] = 0 DO Dec(j);
       last := j;
       row := 0;
       FOR j := 0 TO last DO
         BEGIN
           IF flnbr THEN write(F,Pad2Len(itoa(j),4));
           write(F,Pad2Len(labels[j],11));
           temp := nbr2op(lines[j],banks[j]);
           p := Pos('(',temp);
           IF (p > 0) AND (labels[lines[j] AND $7FF] <> '') THEN Delete(temp,p,5);
           write(F,Pad2Len(temp,26));
           IF comments[j] <> '' THEN writeln(F,';'+ comments[j])
                    ELSE writeln(F);
           Inc(row);
           IF (name = 'con') AND (row > 23) THEN
             BEGIN row := 0; ch := Getkey;
                  IF ch = #27 THEN goto out1;
             END;
         END;
 out1: Close(F);
     END;

{**********************************************************************}

  VAR
    j,col,row,nbr,curline,wstart,regline,regstart : Integer;
    lit,tempop,dir,bits,dest,eestart,datline,code : Integer;
    ch : Char;
    linenbr : String[5];                   { for line numbers (program steps) }
    opstr : Str26;                                         { operation string }
    regstr : String[11];
    switch : Boolean;
  LABEL drawscreen, display, newlit, newopcode, getdest;
{**********************************************************************}
  PROCEDURE nbr2parm(n:Word;b:Boolean);   { change op code to parameters }   
     BEGIN

       CASE (n SHR 12) OF
         3: BEGIN                                 { Literal operations }
           CASE (n SHR 8 AND $0F) OF
            0: BEGIN row := 19; col := 0 END;
            4: BEGIN row := 21; col := 30 END;
            8: BEGIN row := 21; col := 20 END;
            9: BEGIN row := 19; col := 20 END;
            $0A: BEGIN row := 19; col := 40 END;
            $0C: BEGIN row := 21; col := 10 END;
            $0E: BEGIN row := 19; col := 10 END;
           END;
           lit := n AND $0FF;
         END;
        2: BEGIN                                 { goto or call }
            IF (n AND $0800 > 0 ) THEN BEGIN row := 23; col := 0 END
              ELSE BEGIN row := 19; col := 30 END;
            dest := n AND $07FF;
          END;
        1: BEGIN                            { bit oriented }
             bits := n SHR 7 AND 7;
             CASE (n SHR 10 AND 3) OF
               0: BEGIN row := 19; col := 60 END;
               1: BEGIN row := 20; col := 60 END;
               2: BEGIN row := 21; col := 60 END;
               3: BEGIN row := 22; col := 60 END;
             END;  { Case }
             regline := 0; regstart := n AND $7F;
           END;
        0: BEGIN                          { register oriented }
           regline := 0; regstart := n AND $7F;
           CASE (n SHR 8 AND $0F) OF
            0: if ( n SHR 7 AND 1) > 0 THEN BEGIN row := 20; col := 0 END
                    ELSE CASE(n AND $0F) OF
                         0: BEGIN row := 22; col := 0 END;
                         2: BEGIN row := 23; col := 60 END;
                         3: BEGIN row := 23; col := 10 END;
                         4: BEGIN row := 23; col := 20 END;
                         5: BEGIN row := 23; col := 30 END;
                         6: BEGIN row := 23; col := 40 END;
                         8: BEGIN row := 20; col := 30 END;
                         9: BEGIN row := 22; col := 30 END;
                     END;
            1: IF ( n SHR 7 AND 1)> 0 THEN BEGIN row := 20; col := 50 END
                    ELSE BEGIN row := 19; col := 50 END;
            2: BEGIN row := 22; col := 10 END;
            3: BEGIN row := 22; col := 70 END;
            4: BEGIN row := 22; col := 20 END;
            5: BEGIN row := 20; col := 20 END;
            6: BEGIN row := 20; col := 40 END;
            7: BEGIN row := 20; col := 10 END;
            8: BEGIN row := 21; col := 0 END;
            9: BEGIN row := 21; col := 50 END;
            $0A: BEGIN row := 19; col := 70 END;
            $0B: BEGIN row := 22; col := 70 END;
            $0C: BEGIN row := 22; col := 40 END;
            $0D: BEGIN row := 21; col := 40 END;
            $0E: BEGIN row := 22; col := 50 END;
            $0F: BEGIN row := 20; col := 70 END;
           END;
       IF ((n SHR 8 AND $0F) > 1) THEN   { not clrf,clrw,movwf or nop }
            IF(n SHR 7 AND 1) > 0 THEN  dir := 1     { d bit }
               ELSE dir := 0;
       IF n = $65 THEN BEGIN row := 23; col := 30 END;
       IF n = $66 THEN BEGIN row := 23; col := 40 END;
      END;  { 0 }
    END;  { Case }
  END;  { nbr2parm }              

{**********************************************************************}

BEGIN
    FOR j:=0 TO 1023 DO lines[j] := 0;     { set starting values of variables }
    FOR j:=0 TO 63 DO edata[j] := 0;
    FOR j:=0 TO 1023 DO labels[j] := '';
    FOR j:=0 TO 1023 DO comments[j] := '';
    FOR j:=0 TO 1023 DO banks[j] := FALSE;
    col := 0;   dir := 0; bits := 0;
    row := 19;  dest := 0; lit := 0; eestart := 0; datline :=0;
    wstart :=0;  curline := 0; regstart := 0; regline := 0;
    fname := 'NONAME';     ffirst := FALSE;
    IF mode = 7 THEN BEGIN
        Vbase := $B000; Cursor1 := $70; Cursor2 := $70; Cursor3 := $70 END;
  drawscreen:
    ClrScr;                                                { Clear the screen }
    HideCursor;
    FOR j:=0 TO 37 DO 
      fwrite((j MOD 8)*10+1,(j DIV 8)+20,keylabels[j],$7);
    ClearBox(1,1,28,16,$E);                                   { program box }
    fwrite(12,1,'Program',$E);
    fwrite(4,16,'Home/End',$C);
    fwrite(14,16,'Ctrl-Home/End',$D);
    ClearBox(30,1,14,16,$B);                                { registers box }
    fwrite(33,1,'Registers',$B);
    fwrite(32,16,'PgUp/PgDn',$C);
    ClearBox(45,1,33,8,$6);                                      { data box }
    fwrite(59,1,'Data',$6);
    fwrite(54,8,'Ctrl-PgUp/PgDn',$C);
    ClearBox(50,10,28,3,$A);                                  { literal box }
    fwrite(57,10,'literal (k)',$A);                   
    fwrite(57,12,'Alt-nbr/+/-',$A);                   
    ClearBox(45,14,28,3,$C);                                { operation box }
    fwrite(55,14,'operation',$C);
    fwrite(50,16,'EnterInsertDelete',$C);
    ClearBox(74,14,5,3,$B);                               { destination box }
    fwrite(75,16,'Tab',$B);
    ClearBox(45,10,3,3,$B);                                      { bank box }
    fwrite(45,10,'RP0',$B);
     CommentBox;                                     { comment box & label box}
    fwrite(47,9,'BIT 7 6 5 4 3 2 1 0 DEC HEX ASC',$F);
    fwrite(1,25,'F1-help 2-save 3-load 4-list 5-reg ' + 
                '6-label 7-comnt 8-copy 9-line 10-goto 11-oper',$E);
  display:
       { Go through program and set bank switches according to Status bit 5 }
        switch := FALSE;
        FOR j:=0 TO 1023 DO 
        BEGIN
          banks[j] := Switch;                       { mark it before change }
          IF lines[j] = $1683 THEN switch := TRUE;
          IF lines[j] = $1283 THEN switch := FALSE;
        END;
        FOR j:=wstart TO  wstart+13 DO 
           BEGIN
             opstr :=itoa(j)+ ' ' + nbr2op(lines[j],banks[j]);   
             opstr := pad2len(opstr,26);
             fwrite(2,j-wstart+2,opstr,15);              { display operation }
           END;
                                                 { Display data in data box }
        FOR j := eestart TO eestart+5 DO BEGIN
           GotoXy(47,j-eestart+2); Write(j:2,'  ');            { data line # }
           prtbin(edata[j]);                           { display binary data }
         END;
       FOR j:=regstart TO regstart+13 DO BEGIN           { Display Registers }
           regstr := reg[j];
           regstr := pad2len(regstr,11);
           fwrite(31,j-regstart+2,regstr,15);
         END;
      newlit:
        GotoXY(51,11);                                    { Display Literal }
        prtbin(lit);
    newopcode:
       nbr := (row-19)*8+col DIV 10;   { change cursor pos to a opcode number }
       tempop := opcode[nbr];                 { raw opcode without parameters }
       CASE(tempop SHR 12) OF                    { get present operation type }
          3:  tempop := tempop + lit;                 { and add in parameters }
          2:  tempop := tempop + dest;
          1: BEGIN
            tempop := tempop +regstart+regline;
            tempop := tempop OR (bits SHL 7);
            END;
          0:
            if ((tempop SHR 8 AND $0F) > 1) THEN BEGIN
                tempop := tempop + regstart+regline;
                tempop := tempop + (dir SHL 7);
               END
            else if (tempop SHR 7 AND 1 )> 0 THEN
                     tempop := tempop +regstart+regline;
        END;  { Case }
       REPEAT
             chgattr(2,curline+2,26,cursor2);      { cursor line program box }
             chgattr(31,regline+2,12,cursor2);    { cursor line register box }
             chgattr(46,datline+2,31,cursor2);        { cursor line data box }
             chgattr(Succ(col),Succ(row),labelsize[nbr],cursor1);  { opcode cursor }
             chgattr(65-2*bits,9,1,cursor2);            { literal bit cursor }
             IF (dest >= wstart) AND (dest <= wstart +13) THEN
                   chgattr(2,dest-wstart+2,26,cursor3); { destination cursor }
             fwrite(46,15,'                          ',$F);  { off old }
                                                      { put in new operation }
             fwrite(46,15,nbr2op(tempop,banks[wstart+curline]),15);  
                                                         { destination (W/f) }
             GotoXY(75,15); if (dir = 0) THEN Write(' W ')
                  else Write(' f ');
             Fwrite(2,18,Pad2len(labels[wstart+curline],15),$F);     { label }                         { Display Literal }
             Fwrite(39,18,Pad2len(comments[wstart+curline],37),$F); { comment }                         { Display Literal }
             IF banks[wstart + curline] = TRUE THEN fwrite(46,11,'1',$F)
               ELSE fwrite(46,11,'0',$F);                     { write bank # }
             ch := GetKey;                { wait for character from keyboard }
             chgattr(Succ(col),Succ(row),labelsize[nbr],$07);{ op cursor off }
             chgattr(2,curline+2,26,$F);                { program cursor off }
             chgattr(31,regline+2,12,$F);              { register cursor off }
             chgattr(46,datline+2,31,$F);                  { data cursor off }
             chgattr(65-2*bits,9,1,$F);                    { literal bit off }
             IF (dest >= wstart) AND (dest <= wstart +13) THEN
                   chgattr(2,dest-wstart+2,26,$F);      { destination cursor }
             CASE ch OF
                 HOME:                     { Home program cursor up one line }
                  if (curline > 0) THEN                     { cursor at top? }
                        Dec(curline)                        { no, move it up }
                  else                                 { must move window up }
                     if (wstart>0) THEN BEGIN          { if not at beginning }
                        Dec(wstart);
                        goto display;
                       END;
                 ENDKEY:                  { End program cursor down one line }
                  if (curline < 13) THEN                        { at bottom? }
                     Inc(curline)                            { no, move down }
                  else                                    { move window down }
                  if (wstart<1010) THEN BEGIN                { if not at end }
                          Inc(wstart);
                          goto display;
                    END;
                 INS: BEGIN                             { Insert instruction }
                  if (insertproc(wstart+curline))>0  THEN BEGIN
                             { if allowed }
                     IF ((tempop AND $3000) = $2000) AND
                       ((tempop AND $07FF) > wstart+curline) THEN
                                       Inc(tempop); 
                          { if call or goto, inc }
                      lines[wstart+curline] := tempop;
                      END; 
                    goto display; 
                   END;  
                  DEL:  BEGIN                           { delete instruction }
                            deleteproc(wstart+curline);
                            goto display;
                          END; 
              CtrlPgUp:                  { Crtl-Pgup data cursor up one line }
                  if (datline > 0)  THEN                    { cursor at top? }
                        Dec(datline)                        { no, move it up }
                  else                                 { must move window up }
                     if (eestart>0) THEN BEGIN         { if not at beginning }
                        Dec(eestart);
                        goto display;
                       END;
              CtrlPgDn:                { Ctrl-Pgdn data cursor down one line }
                   if (datline < 5) THEN                        { at bottom? }
                      Inc(datline)                           { no, move down }
                  else                                    { move window down }
                  if (eestart < 58) THEN BEGIN               { if not at end }
                          Inc(eestart);
                          goto display;
                    END;
             CtrlEnterKp:  BEGIN             { Ctrl-Enter toggle bit in data }
                    IF (1 SHL bits AND edata[eestart+datline] ) = (1 SHL bits)
                       THEN
                          edata[eestart+datline] := 
                                      edata[eestart+datline]  - (1 SHL bits) 
                      ELSE edata[eestart+datline]  := 
                          edata[eestart+datline]  + (1 SHL bits);  
                      goto display;
                    END;
             CtrlHome: BEGIN                    { Ctrl-Home - decrement dest }
                  IF dest > 0 THEN Dec(dest); 
                  goto display;
                END;
             CtrlEnd: BEGIN                     { Ctrl-End - increment dest  }
                  IF dest < 1023 THEN Inc(dest); 
                  goto display;
                END;       
             CtrlPlusKp: BEGIN                { Ctrl- + KP - increment edata }
                  if (edata[eestart+datline]  < 255) THEN 
                                   Inc(edata[eestart+datline] );
                  goto display;
                END;
             CtrlMinusKp: BEGIN                { Ctrl- -KP - decrement edata }
                     IF (edata[eestart+datline]  > 0) THEN 
                                   Dec(edata[eestart+datline] );
                     goto display;
                  END;          
              #187: BEGIN                         { F1 - display help screen }
                       helpscreen;
                       goto drawscreen; 
                    END; 
              #188: BEGIN                           { F2 - output a hex file }
                      writefile;
                      CommentBox;
                      goto display;
                    END;
              #189: BEGIN                            { F3 - input a hex file }
                      readfile;
                      CommentBox;
                      ffirst := FALSE;
                      goto display;
                    END;
              #190: BEGIN                           { F4 - List to disk etc. }
                      flnbr := FALSE;
             getdest: ch :=
                  Response('List to Disk,Printer,Scrn,LiNe #''s or Cancel? (D,P,S,N,C)');
                      CASE UpCase(ch) OF
                        'D' : BEGIN
                                GetString('Filename: ',8,fname,'.LST');
                                ListFile(fname+'.LST');
                              END;
                        'S' : BEGIN
                                ClrScr; GotoXY(1,1);
                                ListFile('con');
                                ch := Getkey;     { to hold last page }
                                Goto drawscreen;
                              END;
                        'P' : Listfile('prn');
                      'N','L' : BEGIN
                                  flnbr := TRUE; goto getdest END;
                      END;  { Case }
                      CommentBox;
                      goto display;
                    END;
               #191: BEGIN                       { F5 - change register name }  
                      IF(regstart+regline > 11) THEN
                               Chgregname(regstart+regline);
                       goto display;
                    END;
              #192: BEGIN                               { F6 - insert label }  
                       InsertLabel(wstart+curline);
                       goto display;
                    END;                    
              #193: BEGIN                             { F7 - insert comment }  
                       InsertComment(wstart+curline);
                       goto display;
                    END;
              #194: BEGIN                         { F8 - insert instruction } 
                { instruction at cursor is inserted just before dest cursor }
                      tempop := lines[wstart + curline];
                      if (insertproc(dest))>0  THEN BEGIN
                             { if allowed }
                         IF ((tempop AND $3000) = $2000) AND
                            ((tempop AND $07FF) > dest) THEN
                                       Inc(tempop); 
                                 { if call or goto, inc }
                         lines[dest] := tempop;
                         IF dest > wstart + curline THEN j := 1 ELSE j := 2;
                         comments[dest] := comments[wstart+curline+j-1];
                         Inc(dest);
                         if (curline + j < 13) THEN           { at bottom? }
                               Inc(curline,j)              { no, move down }
                           else                         { move window down }
                             if (wstart<1009) THEN         { if not at end }
                               Inc(wstart,j);
                      END;   
                      goto display; 
                    END;
              #195: BEGIN                      { F9 - goto a certain line # }
                      linenbr := '0';
                      GetString('Input line #: ',3,linenbr,'');
                      val(linenbr,j,code);
                      Dec(j,curline);       { to keep curline at same place }
                      if (j > 1010) THEN j := 1010;
                      if (j < 0) THEN j := 0;
                      wstart := j;
                      CommentBox;
                      goto display; 
                    END;
              #196: BEGIN                   { F10 - set destination cursor }
                      linenbr := '0';
                      GetString('Input line #: ',3,linenbr,'');
                      val(linenbr,j,code);
                      if (j > 1010) THEN j := 1010;
                      if (j < 0) THEN j := 0;
                      dest := j;
                      CommentBox;
                      goto display; 
                    END;
           #5,#174: BEGIN                   { F11 - copy to operation window }
                      nbr2parm(lines[wstart + curline],banks[wstart + curline]);
                      goto display; 
                    END;
              PGUP: BEGIN                      { PgUp register cursor up one }
                  if (regline > 0) THEN BEGIN
                        Dec(regline);
                        if (regline+regstart = 7) THEN Dec(regline);
                     END
                  else                          { scroll if not at beginning }
                     if (regstart>0) THEN BEGIN
                        Dec(regstart);
                        goto display;
                       END;
                  goto newopcode;
                  END;
             PGDN: BEGIN                         { PgDn register cursor down }
                   if (regline < 13) THEN BEGIN 
                      Inc(regline);
                      if (regline+regstart = 7) THEN Inc(regline);
                    END
                  else                                              { scroll }
                  if (regstart<66) THEN BEGIN
                          Inc(regstart);
                          goto display;
                    END;
                  goto newopcode;
                  END;
          UARR: BEGIN               { up arrow - operation cursor up a line }
                   if (row > 19) THEN Dec(row);
                   goto newopcode;
                 END;
          DARR: BEGIN           { down arrow - operation cursor down a line }
                    if (row < 23) THEN Inc(row);
                    if((row = 23) AND (col>20))  THEN col := 20;
                    goto newopcode;
                 END;
          RARR:  BEGIN                         { right arrow - cursor right }
                if (col < 70) THEN col:= col + 10;
                if((row = 23) AND (col>50)) THEN col := 50;
                goto newopcode;
                  END;
          LARR:  BEGIN                           { left arrow - cursor left }
                if (col > 0) THEN col:= col -10;
                goto newopcode;
                  END;
                                     { Alt-number keys toggle bits in literal }
             #1: BEGIN                                                { bit 0 }
                   IF (lit AND 1) = 1 THEN lit := lit AND $FE 
                       ELSE lit := lit OR 1;
                   goto newlit;
                 END;
             #2: BEGIN                                            { Alt minus }
                   IF lit > 0 THEN Dec(lit); 
                   goto newlit;
                 END;
             #3: BEGIN                                             { Alt plus }
                   IF lit < 255 THEN Inc(lit); 
                   goto newlit;
                 END;
             #248: BEGIN                                              { bit 1 }
                   IF (lit AND 2) =2 THEN lit := lit AND $FD
                       ELSE lit := lit OR 2;
                   goto newlit;
                 END;
            #249: BEGIN
                   IF (lit AND 4) =4 THEN lit := lit AND $FB
                       ELSE lit := lit OR 4;
                   goto newlit;
                 END;
            #250: BEGIN
                   IF (lit AND 8) =8 THEN lit := lit AND $F7
                       ELSE lit := lit OR 8;
                   goto newlit;
                 END;
            #251: BEGIN
                   IF (lit AND 16) =16 THEN lit := lit AND $EF
                       ELSE lit := lit OR 16;
                   goto newlit;
                 END;
            #252: BEGIN
                   IF (lit AND 32) = 32 THEN lit := lit AND $DF
                       ELSE lit := lit OR 32;
                   goto newlit;
                 END;
            #253: BEGIN
                   IF (lit AND 64) = 64 THEN lit := lit AND $BF
                       ELSE lit := lit OR 64;
                   goto newlit;
                 END;
            #254: BEGIN
                   IF (lit AND 128) = 128 THEN lit := lit AND $7F
                       ELSE lit := lit OR 128;
                   goto newlit;
                 END;
          #9:  BEGIN                        {Tab - toggle destination (W/f) }
                 IF (dir = 0) THEN dir := 1 else dir := 0;
                 goto newopcode;
               END;
          #244: BEGIN                               { Ctrl-Rtarr cursor right }
                if (bits > 0) THEN Dec(bits);
                goto newopcode;
               END;
          #243:  BEGIN                           { Ctrl-Ltarr bit cursor left }
                  if (bits < 7) THEN Inc(bits);
                  goto newopcode;
                END;
         ENTER: BEGIN         { Enter - replace current operation in program }
             lines[wstart+curline] := tempop;
             if(curline < 13) THEN 
                    Inc(curline)
                else
                  if(wstart<1010) THEN
                         Inc(wstart);
             goto display;
             END;      
           END;  { Case }
         UNTIL ch = ESC;                          { until Esc key is pressed }
    ch := Response('Really want to quit? (Y/N)');
    IF UpCase(ch) <> 'Y' THEN BEGIN CommentBox; goto display END;
    code := mode;
    ASM mov ax,code; int 10h END;
 END. 


